mod config;
mod db;

use std::{net::SocketAddr, sync::Arc};
use tokio::{net::{TcpListener, TcpStream}, sync::Mutex};
use crate::config::Config;
use figment::{Figment, providers::{Format, Env, Json}};
use futures_util::{StreamExt, stream::SplitSink, TryStreamExt, SinkExt};
use tokio_tungstenite::{WebSocketStream, tungstenite::Message as TungsteniteMessage};
use common::packets::Message;

type PeerList = Arc<Mutex<Vec<Peer>>>;

pub struct Peer {
    pub addr: SocketAddr,
    pub websocket_tx: SplitSink<WebSocketStream<TcpStream>, TungsteniteMessage>
}

impl Peer {
    fn new(addr: SocketAddr, websocket_tx: SplitSink<WebSocketStream<TcpStream>, TungsteniteMessage>) -> Self {
        Self {
            addr,
            websocket_tx
        }
    }
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    let config: Config = Figment::new()
        .merge(Json::file("server-config.json"))
        .merge(Env::prefixed("APP_"))
        .extract()?;

    let peer_list =  Arc::new(Mutex::new(Vec::new()));
    let tcp_listener = TcpListener::bind(format!("{}:{}", config.host, config.port)).await?;
    println!("Listening on {}:{}", config.host, config.port);

    while let Ok((stream, addr)) = tcp_listener.accept().await {
        println!("Connection from {}", &addr);
        tokio::spawn(handle_connection(stream, addr, peer_list.clone()));
    }

    Ok(())
}

async fn handle_connection(stream: TcpStream, addr: SocketAddr, peer_list: PeerList) {
    let websocket_stream = tokio_tungstenite::accept_async(stream)
        .await
        .expect("Could not initialize websocket stream");

    let (websocket_tx, mut websocket_rx) =  websocket_stream.split();

    {
        let mut locked_peer_list = peer_list.lock().await;
        locked_peer_list.push(Peer::new(addr, websocket_tx));
    }

    while let Some(socket_result) = websocket_rx.next().await {
        match socket_result {
            Ok(websocket_message) => {
                handle_websocket_message(websocket_message, addr, peer_list.clone()).await;
            }
            Err(err) => {
                eprintln!("SocketError: {err}");
            }
        }

    }
}

async fn handle_websocket_message(websocket_message: TungsteniteMessage, addr: SocketAddr, peer_list: PeerList) {
    match websocket_message {
        TungsteniteMessage::Text(text_message) => {
            match serde_json::from_str(&text_message) {
                Ok(message) => {
                    if let Err(err) = handle_message(message, addr, peer_list.clone()).await {
                        eprintln!("Handle message error: {err}");
                    }
                }
                Err(err) => {
                    eprintln!("ParseError: {err}")
                }
            }
        }
        TungsteniteMessage::Close(_) => {
            let mut locked_peer_list = peer_list.lock().await;

            for i in 0..locked_peer_list.len() {
                if locked_peer_list[i].addr.to_string() == addr.to_string() {
                    locked_peer_list.remove(i);
                    break;
                }
            }
        }
        _ => {
            println!("Websocket: Unsupported message type");
        }
    }
}


async fn handle_message(message: Message, sender: SocketAddr, peer_list: PeerList) -> anyhow::Result<()> {
    match &message {
        Message::Text { chat, content} => {
            let mut use_peer_list = peer_list.lock().await;

            for mut peer in &mut *use_peer_list {
                println!("[Message]{}: {content}", peer.addr.to_string());
                if peer.addr.to_string() == sender.to_string() {
                    continue;
                }
                peer.websocket_tx.send(
                    tokio_tungstenite::tungstenite::Message::Text(
                        serde_json::to_string(&message)?
                    )
                ).await?;
            }

        }
    }

    Ok(())
}